/***************************************************************************
 *   Copyright 1998,2000  Stephan Kulow <coolo@kde.org>                    *
 *   Copyright 2008 by Davide Bettio <davide.bettio@kdemail.net>           *
 *   Copyright 2009, 2011 by Glad Deschrijver <glad.deschrijver@gmail.com> *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/

#include "luna.h"

#include <assert.h>

#include <QGraphicsLinearLayout>
#include <QPainter>

#include <KComboBox>
#include <KConfigDialog>
#include <KDebug>
#include <KUrlCompletion>

#include <Plasma/IconWidget>
#include <Plasma/Svg>
#include <Plasma/Theme>
#include <Plasma/ToolTipManager>

#include "lunacalc.h"
#include "lunadialog.h"

Luna::Luna(QObject *parent, const QVariantList &args)
    : Plasma::Applet(parent, args),
      m_counter(-1)
{
	setHasConfigurationInterface(true);
	setAspectRatioMode(Plasma::ConstrainedSquare);
	size = 82;
	resize(QSize(size, size));

	m_lunaCalc = new LunaCalc(this);
	m_lunaDialog = 0;
}

void Luna::setImage(const QString &fileName)
{
	m_theme->setImagePath(fileName);

	if (!m_theme->isValid()) {
		m_theme->setImagePath("widgets/luna-gskbyte");
		if (!m_theme->isValid())
			setFailedToLaunch(true, i18n("The luna SVG file was not found"));
	}
}

void Luna::setShowBackground(bool showBackground)
{
	if (showBackground) {
		setBackgroundHints(Plasma::Applet::StandardBackground);
	} else {
		setBackgroundHints(Plasma::Applet::NoBackground);
	}
}

void Luna::init()
{
	m_theme = new Plasma::Svg(this);
	m_theme->setContainsMultipleImages(true);

	configChanged();

	Plasma::ToolTipManager::self()->registerWidget(this);

	connectToEngine();

	QGraphicsLinearLayout *mainLayout = new QGraphicsLinearLayout;
	mainLayout->setContentsMargins(0, 0, 0, 0);
	mainLayout->setSpacing(0);
	Plasma::IconWidget *mainButton = new Plasma::IconWidget(this);
	connect(mainButton, SIGNAL(clicked()), this, SLOT(showDialog()));
	mainLayout->addItem(mainButton);
	setLayout(mainLayout);
}

Luna::~Luna()
{
	delete m_theme;
	if (m_lunaDialog)
		delete m_lunaDialog;
}

void Luna::configChanged()
{
	m_northHemisphere = config().readEntry("northHemisphere", true);
	m_showBackground = config().readEntry("showBackground", false);
	m_recentImage = config().readEntry("recentImage", "widgets/luna2");
	m_recentImages = config().readEntry("recentImages", QStringList());

	setImage(m_recentImage);
	setShowBackground(m_showBackground);
}

LunaDialog *Luna::dialog()
{
	if (!m_lunaDialog) {
		m_lunaDialog = new LunaDialog();
		m_lunaDialog->resize(QSize(200, 100));
		m_lunaDialog->setDateTime(m_todayDateTime);
	}
	m_lunaDialog->setTodayDateTime(m_todayDateTime);
	return m_lunaDialog;
}

void Luna::showDialog()
{
	dialog();
	if (!m_lunaDialog->isVisible()) {
		m_lunaDialog->show();
		m_lunaDialog->move(popupPosition(m_lunaDialog->size()));
	} else {
		m_lunaDialog->hide();
	}
}

void Luna::constraintsEvent(Plasma::Constraints constraints)
{
	if (constraints & Plasma::SizeConstraint) {
		if (formFactor() == Plasma::Horizontal) {
			setMinimumWidth(geometry().height());
		} else if (formFactor() == Plasma::Vertical) {
			setMinimumHeight((int)geometry().width());
		}
	}
	setShowBackground(m_showBackground); // make sure that the background is hidden on KDE restart if necessary
}

void Luna::connectToEngine()
{
	Plasma::DataEngine* timeEngine = dataEngine("time");
	timeEngine->connectSource("Local", this, 360000, Plasma::AlignToHour);
}

void Luna::dataUpdated(const QString& source, const Plasma::DataEngine::Data &data)
{
	Q_UNUSED(source)

	QDateTime dt(data["Date"].toDate(), data["Time"].toTime());
	calcStatus(dt.toTime_t());

	m_todayDateTime = dt;
}

/***************************************************************************/

void Luna::createConfigurationInterface(KConfigDialog *parent)
{
	QWidget *widget = new QWidget();
	ui.setupUi(widget);
	connect(parent, SIGNAL(applyClicked()), this, SLOT(configAccepted()));
	connect(parent, SIGNAL(okClicked()), this, SLOT(configAccepted()));
	parent->addPage(widget, i18nc("@title:tab General configuration", "General"), icon());

	ui.northernRadio->setChecked(m_northHemisphere);
	ui.southernRadio->setChecked(!m_northHemisphere);

	ui.showBackground->setChecked(m_showBackground);

	ui.imageUrlRequester->comboBox()->setEditable(true);
	ui.imageUrlRequester->comboBox()->setMinimumContentsLength(10);
	ui.imageUrlRequester->comboBox()->setSizeAdjustPolicy(QComboBox::AdjustToMinimumContentsLengthWithIcon);
	ui.imageUrlRequester->comboBox()->setCompletionObject(new KUrlCompletion());
	ui.imageUrlRequester->comboBox()->setAutoDeleteCompletionObject(true);
	ui.imageUrlRequester->setMode(KFile::File | KFile::ExistingOnly);
	ui.imageUrlRequester->setFilter("*.svg *.svgz|" + i18nc("@item:inlistbox filter", "SVG files") + "\n*|" + i18nc("@item:inlistbox filter", "All files"));
	connect(ui.imageUrlRequester, SIGNAL(returnPressed(QString)),
	        this, SLOT(setFileName(QString)));
	connect(ui.imageUrlRequester, SIGNAL(urlSelected(KUrl)),
	        this, SLOT(setUrl(KUrl)));
	ui.imageUrlRequester->comboBox()->addItems(m_recentImages);
	ui.imageUrlRequester->setUrl(KUrl::fromPath(m_recentImage));
	ui.imageUrlRequester->comboBox()->setCurrentIndex(ui.imageUrlRequester->comboBox()->findText(m_recentImage));

	connect(ui.northernRadio, SIGNAL(clicked(bool)), parent, SLOT(settingsModified()));
	connect(ui.southernRadio, SIGNAL(clicked(bool)), parent, SLOT(settingsModified()));
	connect(ui.showBackground, SIGNAL(clicked(bool)), parent, SLOT(settingsModified()));
	connect(ui.imageUrlRequester, SIGNAL(textChanged(QString)), parent, SLOT(settingsModified()));
}

void Luna::configAccepted()
{
	m_northHemisphere = ui.northernRadio->isChecked();
	m_showBackground = ui.showBackground->isChecked();
	m_recentImage = ui.imageUrlRequester->url().pathOrUrl();
	m_recentImages.clear();
	for (int i = 0; i < ui.imageUrlRequester->comboBox()->count(); ++i)
		m_recentImages << ui.imageUrlRequester->comboBox()->itemText(i);

	config().writeEntry("northHemisphere", m_northHemisphere);
	config().writeEntry("showBackground", m_showBackground);
	config().writeEntry("recentImage", m_recentImage);
	config().writeEntry("recentImages", m_recentImages);

//	setImage(m_recentImage);
//	setShowBackground(m_showBackground);

	update();

	emit configNeedsSaving();
}

void Luna::paintInterface(QPainter *p, const QStyleOptionGraphicsItem *option, const QRect &contentsRect)
{
	Q_UNUSED(option)

	if (!m_theme) {
		return;
	}

	if (m_northHemisphere) {
		m_theme->paint(p, contentsRect, QString::number(m_counter));
	} else {
		p->save();
		p->rotate(180);
		p->translate(-geometry().width(), -geometry().height());
		m_theme->paint(p, contentsRect, QString::number(m_counter));
		p->restore();
	}
}

/***************************************************************************/

void Luna::calcStatus(time_t time)
{
	Plasma::ToolTipContent toolTipData;

	QList<QDateTime> phases = m_lunaCalc->calcStatus(time);
	QDateTime ln = phases.at(0);
	kDebug() << "last new " << KGlobal::locale()->formatDateTime(ln);

	QDateTime fq = phases.at(1);
	kDebug() << "first quarter " << KGlobal::locale()->formatDateTime(fq);

	QDateTime fm = phases.at(2);
	kDebug() << "full moon " << KGlobal::locale()->formatDateTime(fm);

	QDateTime tq = phases.at(3);
	kDebug() << "third quarter " << KGlobal::locale()->formatDateTime(tq);

	QDateTime nn = phases.at(4);
	kDebug() << "next new " << KGlobal::locale()->formatDateTime(nn);

	QDateTime now;
	now.setTime_t(time);
	kDebug() << "now " << KGlobal::locale()->formatDateTime(now);

	m_counter = ln.daysTo(now);
	kDebug() << "counter " << m_counter << " " << fm.daysTo(now);

	if (fm.daysTo(now) == 0) {
		m_counter = 14;
		toolTipData.setMainText(i18nc("@info:tooltip", "Full Moon"));
	} else if (m_counter <= 15 && m_counter >= 13) {
		m_counter = 14 + fm.daysTo(now);
		kDebug() << "around full moon " << m_counter;
	}

	int diff = fq.daysTo(now);
	if (diff == 0)
		m_counter = 7;
	else if (m_counter <= 8 && m_counter >= 6) {
		m_counter = 7 + diff;
		kDebug() << "around first quarter " << m_counter;
	}

	diff = ln.daysTo(now);
	if (diff == 0)
		m_counter = 0;
	else if (m_counter <= 1 || m_counter >= 28) {
		m_counter = (29 + diff) % 29;
		diff = -nn.daysTo(now);
		if (diff == 0)
			m_counter = 0;
		else if (diff < 3)
			m_counter = 29 - diff;
		kDebug() << "around new " << m_counter << " " << diff;
	}

	if (tq.daysTo(now) == 0)
		m_counter = 21;
	else if (m_counter <= 22 && m_counter >= 20) {
		m_counter = 21 + tq.daysTo(now);
		kDebug() << "around third quarter " << m_counter;
	}

	kDebug() << "counter " << m_counter;

	assert (m_counter >= 0 && m_counter < 29);

	switch (m_counter) {
	case 0:
		toolTipData.setMainText(i18nc("@info:tooltip", "New Moon"));
		break;

	case 1:
	case 2:
	case 3:
	case 4:
	case 5:
	case 6:
		toolTipData.setMainText(i18ncp("@info:tooltip", "Waxing Crescent (New Moon was yesterday)", "Waxing Crescent (%1 days since New Moon)", m_counter));
		break;

	case 7:
		toolTipData.setMainText(i18nc("@info:tooltip", "First Quarter"));
		break;

	case 8:
	case 9:
	case 10:
	case 11:
	case 12:
	case 13:
		toolTipData.setMainText(i18ncp("@info:tooltip", "Waxing Gibbous (Tomorrow is Full Moon)", "Waxing Gibbous (%1 days to Full Moon)", -fm.daysTo(now)));
		break;

	case 14:
		toolTipData.setMainText(i18nc("@info:tooltip", "Full Moon"));
		break;

	case 15:
	case 16:
	case 17:
	case 18:
	case 19:
	case 20:
		toolTipData.setMainText(i18ncp("@info:tooltip", "Waning Gibbous (Yesterday was Full Moon)", "Waning Gibbous (%1 days since Full Moon)", fm.daysTo(now)));
		break;

	case 21:
		toolTipData.setMainText(i18nc("@info:tooltip", "Last Quarter"));
		break;

	case 22:
	case 23:
	case 24:
	case 25:
	case 26:
	case 27:
	case 28:
		kDebug() << "nn.days " << ln.daysTo(now) << " " << nn.daysTo(now);
		toolTipData.setMainText(i18ncp("@info:tooltip", "Waning Crescent (Tomorrow is New Moon)", "Waning Crescent (%1 days to New Moon)", -nn.daysTo(now)));
		break;

	default:
		kFatal() << "coolo can't count\n";
	}

	Plasma::ToolTipManager::self()->setContent(this, toolTipData);
	update();
}

/***************************************************************************/

void Luna::addFileNameToList(const QString &fileName)
{
	const int index = ui.imageUrlRequester->comboBox()->findText(fileName);
	if (index >= 0) // then remove item in order to re-add it at the top
		ui.imageUrlRequester->comboBox()->removeItem(index);
	ui.imageUrlRequester->comboBox()->insertItem(0, fileName);
	ui.imageUrlRequester->comboBox()->setCurrentIndex(0);
}

void Luna::setUrl(const KUrl &url)
{
	if (url.isLocalFile())
		addFileNameToList(url.path());
	else
		addFileNameToList(url.prettyUrl());
}

void Luna::setFileName(const QString &fileName)
{
	addFileNameToList(fileName);
}

#include "luna.moc"
